/*
  smtppipe - sends mail from stdin to a smtp relay
  see also: http://www.ogris.de/smtppipe/

  smtppipe --help gives you detailed help
  or look at usage() below

  this software is freely distributable
  as described by the terms of the GNU GPL;
  see http://www.gnu.org/licenses/gpl.html
*/

#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <resolv.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <time.h>
#include "version.h"

#define STDIN 0
#define STDOUT 1
#define STDERR 2

#define BLKSIZE 4096       /* buffers are allocated in multiples of BLKSIZE */
#define SMTP_PORT 25              /* maybe overwritten by cmd line argument */
#define NULLSTRING { NULL, 0, 0 } /* initialization for Strings             */
#define CONNECT_TIMEOUT 60        /* maybe overwritten by cmd line argument */
#define READ_TIMEOUT 120          /* "                                      */
#define WRITE_TIMEOUT 120         /* "                                      */

#define INFOP(a,b) if (verbosity >= 1) outp(STDOUT, a, b)
#define DEBUGP(a,b) if (verbosity >= 2) outp(STDOUT, a, b); \
  if (verbosity >= 3) statm()


/* for generating Message-ID and Date */
char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};


typedef struct {
  char *s;   /* zero terminated                                  */
  size_t l;  /* length of string hold in s                       */
  size_t _l; /* length of buffer allocated; multiples of BLKSIZE */
} String;

struct tm_str {        /* used for generating msg-id and date */
  char tm_str_sec[2];  /* "00".."59"                          */
  char tm_str_min[2];  /* "00".."59"                          */
  char tm_str_hour[2]; /* "00".."23"                          */
  char tm_str_day[2];  /* "01".."31"                          */
  char tm_str_mon[2];  /* "01".."12"                          */
  char tm_str_year[4]; /* "1900".."2038"                      */
  char *tm_str_smon;   /* pointer into months[]               */
  char *tm_str_sday;   /* pointer into days[]                 */
};

enum recipient_determination { /* send email to recipients given by...    */
  ARGUMENT_HEADER_IF_NONE,     /* ...arguments, or header if no arguments */
  ARGUMENT_ONLY,               /* ...arguments only                       */
  ARGUMENT_AND_HEADER,         /* ...arguments and header                 */
  HEADER_ONLY,                 /* ...header only                          */
};


char *relay;         /* signal handlers use it                      */
int read_timeout,    /* we have a generic function called w2s which */
    write_timeout,   /*   we use when talking to the smtp relay     */
    verbosity;       /* 0=quite, 1=info, 2=debug, 3=mem usage       */


/* "initialize" s by assigning NULLSTRING before calling str_realloc */
void str_realloc (String *s, const size_t new_strlen)
{
  static size_t new_buflen;

  new_buflen = (new_strlen / BLKSIZE) * BLKSIZE + BLKSIZE;
  if (new_buflen > s->_l) {
    s->s = realloc(s->s, new_buflen);
    if (!s->s) { write(STDERR, "Fatal error: no memory!\n", 24); exit(2); }
    s->_l = new_buflen;
  }
}

/* append a zero terminated string of known lenght */
void str_cat_n (String *dst, const char *s, const size_t tmp)
{
  static size_t new_strlen;

  new_strlen = dst->l + tmp;
  str_realloc(dst, new_strlen);
  memmove(dst->s + dst->l, s, tmp);
  *(dst->s + new_strlen) = '\0';
  dst->l = new_strlen;
}

/* append a zero terminated string */
void str_cat (String *dst, const char *s)
{
  str_cat_n(dst, s, strlen(s));
}

/* convert unsigned long of known lenght into string */
void ul2str (char *dst, unsigned long int src, int len)
{
  for (--len; len >= 0; --len) {
    dst[len] = '0' + src % 10;
    src /= 10;
  }
}

/* append an unsigned long */
void str_cat_ul (String *dst, unsigned long int i)
{
  static size_t len;
  unsigned long int save;

  save = i;
  len = 0; 
  do {
    ++len;
    i /= 10;
  } while (i);
  str_realloc(dst, dst->l + len);
  *(dst->s + dst->l + len) = '\0';
  ul2str(dst->s + dst->l, save, len);
  dst->l += len;
}

/* insert a character at position pos (pos=0..s.l-1) */
void str_insert_chr (String *s, size_t pos, char c)
{
  str_realloc(s, s->l + 1);
  memmove(s->s + pos + 1, s->s + pos, s->l - pos);
  *(s->s + pos) = c;
  ++s->l;
}

/* return a struct tm_str containing current date and time as strings */
struct tm_str *get_now ()
{
  static time_t sec_epoch = 0;
  static struct tm_str ret;
  static struct tm *tmp;

  if (sec_epoch) return &ret;
  sec_epoch = time(NULL);
  tmp = gmtime(&sec_epoch);
  ret.tm_str_smon = months[tmp->tm_mon];
  ret.tm_str_sday = days[tmp->tm_wday];
  ++tmp->tm_mon;
  tmp->tm_year += 1900;
  ul2str(ret.tm_str_sec, tmp->tm_sec, 2);
  ul2str(ret.tm_str_min, tmp->tm_min, 2);
  ul2str(ret.tm_str_hour, tmp->tm_hour, 2);
  ul2str(ret.tm_str_day, tmp->tm_mday, 2);
  ul2str(ret.tm_str_mon, tmp->tm_mon, 2);
  ul2str(ret.tm_str_year, tmp->tm_year, 4);
  return &ret;
}

/* print s1 to descriptor d; if not NULL print s2 prepended by ": " too */
void outp (const int d, const char *s1, const char *s2)
{
  static String s = NULLSTRING;

  str_cat(&s, s1);

  if (s2) {
    str_cat_n(&s, ": ", 2);
    str_cat(&s, s2);
  }
  str_cat_n(&s, "\n", 1);
  write(d, s.s, s.l);
  s.l = 0;
}

/* inform user on stderr about non fatal error */
void warn (const char *s1, const char *s2)
{
  outp(STDERR, s1, s2);
}

/* print out fatal error and quit program */
void die (const char *s1, const char *s2)
{
  outp(STDERR, s1, s2);
  exit(1);
}

/* show mem usage (call smtppipe with -vvv) */
#include <sys/shm.h>
void statm ()
{
  static int fh, i, len, x;/*i=index of s[], len=len of buf, x for nice output*/
  static char buf[BLKSIZE], *p;
  static ssize_t ret;
  static unsigned long s[7];

  fh = open("/proc/self/statm", O_RDONLY);
  if (fh < 0) return;
  ret = read(fh, buf, BLKSIZE - 1);
  close(fh);
  if (ret <= 0) return;
  for (i = 0, p = buf; i <= 6; ++i, ++p) {
    s[i] = (atol(p) * PAGE_SIZE) >> 10;
    while ((*p >= '0') && (*p <= '9')) ++p;
    if (!*p) break;
  }
  i = 0;
  len = 0;

#define Q(a) \
  memmove(&buf[len], a, strlen(a)); \
  len += strlen(a); \
  memmove(&buf[len], ":", 1); \
  ++len; \
  ul2str(&buf[len], s[i], 20); \
  for (x = 0; (x < 19) && (buf[len + x] == '0'); ++x) buf[len + x] = ' '; \
  len += x - 8 - strlen(a); \
  ul2str(&buf[len], s[i++], 20 - x); \
  len += 20 - x; \
  memmove(&buf[len], " kB\n", 4); \
  len += 4;

  Q("size")
  Q("resident")
  Q("shared")
  Q("code")
  Q("data")
  Q("library")
  Q("dirty")

  write(STDOUT, buf, len);
}

/* generic help */
void usage ()
{
#define HELP \
"smtppipe "VERSION" - reads mail from stdin and forwards it to a smtp relay\n" \
"conformal with sendmail and qmail-inject command line switches\n\n" \
"smtppipe [-a] [-h] [-t] [-A] [-H] [-f sender] [-m smtprelay] [-p port]\n" \
"         [-ct connect_timeout] [-rt read_timeout] [-wt write_timeout] [-v]\n" \
"         [-oem] [-oi] [--] [recipient ... ]\n\n" \
"smtppipe --help\n" \
"smtppipe -help\n" \
"smtppipe -V\n\n" \
"  -a         send message to all addresses given as recipients arguments;\n" \
"             do not use any header recipient addresses\n" \
"  -h, -t     send message to all header recipient addresses (To, Cc, Bcc);\n" \
"             do not use any command line recipients\n" \
"  -A         send message to all addresses given as recipient arguments; if\n" \
"             no recipient arguments are given, use header recipient addresses;\n" \
"             this is the default behaviour\n" \
"  -H         send message to all addresses given as recipient arguments and\n" \
"             to all addresses given as header recipients\n" \
"  -f         forces envelope sender address; if not specified then the\n" \
"             From:-line found in the header, the environment variables USER,\n" \
"             QMAILUSER, LOGNAME and the name of the account under which \n" \
"             smtppipe runs (in that order) are checked for an user name\n" \
"  -m         host name or address of a smtp relay; if not specified then the\n" \
"             environment variables SMTPRELAY, QMAILHOST and MAILHOST (in that\n" \
"             order) are checked for a host name or an address\n" \
"  -p port    use specified port number when connecting to the smtp relay\n" \
"             instead of the default SMTP port; if the host name of the smtp\n" \
"             relay is given in a host:port notation, then that port is used\n" \
"  -ct #      timeout on connecting to smtp relay; defaults to 60 seconds\n" \
"  -rt #      timeout on reading from smtp relay; defaults to 120 seconds\n" \
"  -wt #      timeout on writing to smtp relay; defaults to 120 seconds\n" \
"  -v         be verbose (print out what is done); use -vv for debugging\n" \
"  -oem       not handled due to compatibility\n" \
"  -oi        not handled due to compatibility\n" \
"  --         not handled due to compatibility\n" \
"  recipient  list of space seperated recipient addresses\n" \
"  --help     displays this help\n" \
"  -help      displays this help\n" \
"  -V         displays version\n\n" \
"smtppipe inserts missing From, Subject, Message-ID and Date lines into the\n" \
"header of your mail. Bcc'ed recipients are handled and deleted before final\n" \
"delivery.\n\n" \
"see http://www.ogris.de/smtppipe/\n"

  write(STDOUT, HELP, strlen(HELP));
  exit(0);
}

/* signal handlers used when talking to the smtp relay */
void read_timeout_handler () { die(relay, "read timeout"); }
void write_timeout_handler () { die(relay, "write timeout"); }
void connect_timeout_handler () { die(relay, "connect timeout"); }

/* set the specified signal handler */
void set_alarm (int timeout, void *handler)
{
  static struct sigaction sa;

  memset(&sa, 0, sizeof(struct sigaction));
  if (handler) sa.sa_handler = handler;
  else sa.sa_handler = SIG_DFL;
  if (sigaction(SIGALRM, &sa, NULL)) die("sigaction()", strerror(errno));
  alarm(timeout);
}

/* write command to socket; if specified write value prepended by ": " too    */
/* if specified check whether smtp relay returns the desired code; die if not */
void w2s (int socket, char *command, char *value, char *code)
{
  static String s = NULLSTRING;
  static char buf[BLKSIZE];
  static ssize_t ret;

  if (command) {
    str_cat(&s, command);

    if (value) {
      if (*value != '<') str_cat_n(&s, " ", 1);
      str_cat(&s, value);
    }

    str_cat_n(&s, "\r\n", 2);

    DEBUGP("To smtp relay", s.s);
    set_alarm(write_timeout, write_timeout_handler);
    ret = write(socket, s.s, s.l);
    set_alarm(0, NULL);

    if (ret < 0) die("write(socket)", strerror(errno));
    if ((size_t) ret != s.l)
      die("write(socket)", "could not write whole data stream");
  }

  if (code) {
    set_alarm(read_timeout, read_timeout_handler);
    ret = read(socket, buf, BLKSIZE - 1);
    set_alarm(0, NULL);
    if (ret < 0) die("read(socket)", strerror(errno));
    buf[ret] = '\0';
    DEBUGP("From smtp relay", buf);
    if (strncmp(buf, code, strlen(code))) {
      /* smtp relay did not return our desired status code => quit program */
      String err = NULLSTRING;

      while ((*(buf + ret - 1) == '\r') || (*(buf + ret - 1) == '\n'))
        --ret;

      str_cat(&err, "The smtp relay returns an unexpected status code:\n");
      str_cat_n(&err, buf, ret);
      if (command) {
        str_cat(&err, "\nThe requested command was: ");
        str_cat_n(&err, s.s, s.l - 2);
      }
      str_cat(&err, "\nThe expected status code was");
      die(err.s, code);
    }
  }
  s.l = 0;
}

/* try to get a host name or an ip address of the smtp relay; else die */
/* this can be easily extended to fit your environment                 */
char *get_relay ()
{
  char *relay;

  relay = getenv("SMTPRELAY");
  if (!relay) relay = getenv("QMAILHOST");
  else DEBUGP("SMTPRELAY", relay);
  if (!relay) relay = getenv("MAILHOST");
  else DEBUGP("QMAILHOST", relay);
  if (!relay) die("no smtprelay specified", NULL);
  DEBUGP("MAILHOST", relay);
  return relay;
}

/* try to get a user name which can be used as sender; else die */
/* this can be easily extended to fit your environment          */
char *get_user (const char *hostname)
{
  String s = NULLSTRING;
  struct passwd *pw;
  char *user;

  user = getenv("USER");
  if (!user) user = getenv("QMAILUSER");
  else DEBUGP("USER", user);
  if (!user) user = getenv("LOGNAME");
  else DEBUGP("QMAILUSER", user);
  if (!user) {
    pw = getpwuid(getuid());
    if (pw) user = pw->pw_name;
  }
  else DEBUGP("LOGNAME", user);
  if (!user) die("can not figure out your sender's address", NULL);
  DEBUGP("getpwuid()", user);
  str_cat(&s, user);
  str_cat_n(&s, "@", 1);
  str_cat(&s, hostname);
  return s.s;
}

/* construct message-id from date, time and process id and write it to socket */
void msgid2socket (int socket, char *hostname)
{
  /* Message-Id: <20030811090830.14846.smtppipe@HOSTNAME> */
  static String msgid = NULLSTRING;
  static struct tm_str *now;

  now = get_now();
  str_cat_n(&msgid, "Message-ID: <", 13);
  str_cat_n(&msgid, now->tm_str_year, 4);
  str_cat_n(&msgid, now->tm_str_mon, 2);
  str_cat_n(&msgid, now->tm_str_day, 2);
  str_cat_n(&msgid, now->tm_str_hour, 2);
  str_cat_n(&msgid, now->tm_str_min, 2);
  str_cat_n(&msgid, now->tm_str_sec, 2);
  str_cat_n(&msgid, ".", 1);
  str_cat_ul(&msgid, getpid());
  str_cat_n(&msgid, ".smtppipe@", 10);
  str_cat(&msgid, hostname);
  str_cat_n(&msgid, ">", 1);
  INFOP("Generated Message-ID", msgid.s);
  w2s(socket, msgid.s, NULL, NULL);
}

/* write current time and date to socket */
void date2socket (int socket)
{
  /* Date: Wed, 11 Aug 2003 09:41:30 +0000 */
  static String date = NULLSTRING;
  static struct tm_str *now;

  now = get_now();
  str_cat_n(&date, "Date: ", 6);
  str_cat_n(&date, now->tm_str_sday, 3);
  str_cat_n(&date, ", ", 2);
  str_cat_n(&date, now->tm_str_day, 2);
  str_cat_n(&date, " ", 1);
  str_cat_n(&date, now->tm_str_smon, 3);
  str_cat_n(&date, " ", 1);
  str_cat_n(&date, now->tm_str_year, 4);
  str_cat_n(&date, " ", 1);
  str_cat_n(&date, now->tm_str_hour, 2);
  str_cat_n(&date, ":", 1);
  str_cat_n(&date, now->tm_str_min, 2);
  str_cat_n(&date, ":", 1);
  str_cat_n(&date, now->tm_str_sec, 2);
  str_cat_n(&date, " +0000", 6);
  INFOP("Generated Date", date.s);
  w2s(socket, date.s, NULL, NULL);
}

/* smtp envelope: write address of sender to socket */
void from2socket (int socket, char *from)
{
  INFOP("From", from);
  w2s(socket, "MAIL FROM:", from, "250");
}

/* smtp envelope: write address of a recipient to socket */
void rcpt2socket (int socket, char *rcpt)
{
  INFOP("Recipient", rcpt);
  w2s(socket, "RCPT TO:", rcpt, "250");
}

/* parse an address block (From:, To:, Cc:, Bcc:) in the mail header and */
/*   search for valid mail addresses;                                    */
/* die if mail header contains two or more address blocks of same type   */
/* if not parsing a From: line then return if mode permits using         */
/*   header recipients                                                   */
/* die if there fatal typos (missing brackets or quote signs) or extra   */
/*   whitespace between addresses sperated by commas                     */
void handle_address_line (int socket, char *left, char *right,
                          enum recipient_determination mode,
                          int first_cmd_line_recip,
                          char **from, char *hostname,
                          int *header_this_seen, char *this)
{
  static char *addrleft, *addrright, save;

  if (*header_this_seen)
    die(this, "must not appear twice times within a header!");
  ++*header_this_seen;

  /* check if we have to handle this */
  if ( ((*this != 'F') && ((mode == ARGUMENT_ONLY) ||
        ((mode == ARGUMENT_HEADER_IF_NONE) && (first_cmd_line_recip > 0)))) ||
       ((*this == 'F') && (*from)) )  {
    DEBUGP("handle_address_line called but not needed; line type", this);
    return;
  }

  for (addrleft = left; *addrleft; addrleft = addrright) {

    /* strip whitespace */
    while ((*addrleft == ' ') || (*addrleft == '\r') ||
           (*addrleft == '\t') || (*addrleft == '\n')) ++addrleft;

    if (*addrleft == '"') {
      DEBUGP(addrleft, " is like \"Firstname Surname\" user@domain");
      do {
        ++addrleft;
        if (addrleft >= right) die("unmatched quote sign", left);
      } while ((*addrleft != '"') || (*(addrleft - 1) == '\\'));
      addrleft += 2;
    }
    if (addrleft >= right) die("line contains no address", left);
    addrright = addrleft;
    do {
      ++addrright;
    } while ((*addrright != ' ') && (*addrright != '\r') && (*addrright) &&
             (strncmp(addrright, ", ", 2)) &&   /* commas are allowed */
             (strncmp(addrright, ",\r", 2)));   /*  in mail addresses */
    save = *addrright;
    *addrright = '\0';
    if (*this == 'F') {
      if (*from) die("your mail contains at least two sender addresses", left);
      *from = addrleft;
      from2socket(socket, *from);
    }
    else {
      /* smtp requires a sender before any recipient */
      if (!*from) {
        *from = get_user(hostname);
        from2socket(socket, *from);
      }
      DEBUGP(this, addrleft);
      rcpt2socket(socket, addrleft);
    }
    *addrright = save;
    if (save == ' ') {
      DEBUGP(addrleft, " is like user@domain (Firstname Surname)");
      do {
        ++addrright;
        if (addrright >= right) die("unmachted bracket", addrleft);
      } while ((*addrright != ')') || (*(addrright - 1) == '\\'));
      ++addrright;
    }
    switch (*addrright) {
      case '\0': /* right is reached, stop processing */
        return;
      case '\r': /* continuation on next line guarented by main() */
      case ',':  /* next address follows on this line             */
        ++addrright;
        break;
      default:   /* illegal token */
        die("illegal character found in address line", addrleft);
    } /* switch */
  } /* for */
}

/* if cmd line argument at position pos matches test, then set set_on_found */
/*   to the value of that cmd line argument                                 */
int check_arg (char **argv, int *pos, char *test, char **set_on_found)
{
  static int pos_save;
  static size_t len;

  len = strlen(test);
  if (!strncmp(argv[*pos], test, len)) {
    pos_save = *pos;
    if (strlen(argv[*pos]) > len) *set_on_found = argv[*pos] + len;
    else *set_on_found = argv[++*pos];
    if (!*set_on_found || (**set_on_found == '-'))
      die("invalid parameter", argv[pos_save]);
    return 1;
  }
  return 0;
}

/* main routine, access environment via getenv() */
int main (int argc, char *argv[])
{
  struct hostent *he;
  struct sockaddr_in sin;
  size_t len;
  ssize_t ret;
  enum recipient_determination mode;
  int i, sock, first_cmd_line_recip, connect_timeout,
      header_from_seen, header_to_seen, header_cc_seen, header_bcc_seen,
      header_date_seen, header_msgid_seen, header_subject_seen;
  char *p, *sep, *left, *right, *from;
  char buf[BLKSIZE];
  String mail = NULLSTRING;
  unsigned short port;

  /* initilization */
  first_cmd_line_recip = 0;
  header_from_seen = 0;
  header_to_seen = 0;
  header_cc_seen = 0;
  header_bcc_seen = 0;
  header_date_seen = 0;
  header_msgid_seen = 0;
  header_subject_seen = 0;
  mode = ARGUMENT_HEADER_IF_NONE;
  from = NULL;
  relay = NULL;
  port = SMTP_PORT;
  connect_timeout = CONNECT_TIMEOUT;
  read_timeout = READ_TIMEOUT;
  write_timeout = WRITE_TIMEOUT;
  verbosity = 0;

  /* parse command line arguments */
  for (i = 1; i < argc; ++i) {
    if ( (!strncmp(argv[i], "--help", 6)) || (!strncmp(argv[i], "-help", 5)) )
      usage();
    if (!strncmp(argv[i], "-V", 2)) die("smtppipe " VERSION, NULL);

    if (check_arg(argv, &i, "-f", &from)) continue;   /* from                */
    if (check_arg(argv, &i, "-m", &relay)) continue;  /* relay               */
    if (strncmp(argv[i], "-", 1)) {
      /* if argument does not begin with a '-' it is assumed to be an address */
      first_cmd_line_recip = i;
      break;
    }

    if (!strncmp(argv[i], "--", 2)) continue;         /* mutt                */
    if (!strncmp(argv[i], "-oem", 4)) continue;       /* emacs uses -oem -oi */
    if (!strncmp(argv[i], "-oi", 3)) continue;        /* emacs uses -oem -oi */

    if (check_arg(argv, &i, "-p", &p)) {              /* port                */
      port = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-ct", &p)) {             /* connect timeout     */
      connect_timeout = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-rt", &p)) {             /* read timeout        */
      read_timeout = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-wt", &p)) {             /* write timeout       */
      write_timeout = atoi(p);
      continue;
    }

    if (!strncmp(argv[i], "-a", 2)) {
      mode = ARGUMENT_ONLY;
      continue;
    }

    if ( (!strncmp(argv[i], "-h", 2)) || (!strncmp(argv[i], "-t", 2)) ) {
      mode = HEADER_ONLY;
      continue;
    }

    if (!strncmp(argv[i], "-A", 2)) {
      mode = ARGUMENT_HEADER_IF_NONE;
      continue;
    }

    if (!strncmp(argv[i], "-H", 2)) {
      mode = ARGUMENT_AND_HEADER;
      continue;
    }

    if (!strncmp(argv[i], "-v", 2)) {
      verbosity = strlen(argv[i]) - 1;
      continue;
    }

    warn("unknown parameter", argv[i]);
  } /* parsing cmd line arguments */

  /* must have an smtprelay */
  if (!relay) relay = get_relay();
  DEBUGP("Relay", relay);
  p = strchr(relay, ':');
  if (p) {
    *p++ = '\0';
    port = atoi(p);
  }
  if (verbosity >= 2) {
    String foo = NULLSTRING;
    str_cat_ul(&foo, port);
    DEBUGP("Port", foo.s);
  }
  if (!(he = gethostbyname(relay))) die(relay, hstrerror(h_errno));
  INFOP("FQDN of smtp relay", he->h_name);

  memcpy((char*) &sin.sin_addr, he->h_addr, he->h_length);
  sin.sin_port = htons(port);
  sin.sin_family = PF_INET;

  /* determine local hostname to do the initial SMTP HELO and to form */
  /* the sender's address if not specified otherwise                  */
  if (gethostname(buf, BLKSIZE)) die("gethostname()", strerror(errno));
  DEBUGP("My hostname", buf);
  if (!(he = gethostbyname(buf))) die(buf, hstrerror(h_errno));
  INFOP("My FQDN", he->h_name);

  /* get the mail through STDIN */
  for (;;) {
    ret = read(STDIN, buf, BLKSIZE);
    if (!ret) break;
    if (ret < 0) die("read(STDIN)", strerror(errno));
    str_cat_n(&mail, buf, ret);
  }

  /* count newline w/o prepending '\r' */
  for (len = 0, p = mail.s + 1; *p; ++p) {
    if (*p != '\n') continue;
    if (*(p - 1) == '\r') continue;
    ++len;
  }
  str_realloc(&mail, mail.l + len);
  p = mail.s + mail.l - 1;
  mail.l += len;
  /* replace each single '\n' by "\r\n" */
  for (i = 1; (p > mail.s) && len; --p, ++i) {
    if (*p != '\n') continue;
    if (*(p - 1) == '\r') continue;
    memmove(p + len, p, i);
    i = 0;
    --len;
    *(p + len) = '\r';
  }

  /* no incomplete last line */
  if (*(mail.s + mail.l - 1) ==  '\r') str_cat_n(&mail, "\n", 1);
  else if (strcmp(mail.s + mail.l - 2, "\r\n")) str_cat_n(&mail, "\r\n", 2);

  /* escape single dots on a line */
  for (len = 0; len < mail.l; ++len) {
    p = mail.s + len;
    if (strncmp(p, "\n.\r", 3)) continue;
    len += 3;
    str_insert_chr(&mail, len, '.');
  }

  /* try to find a body seperation */
  sep = strstr(mail.s, "\r\n\r\n");
  if (!sep) DEBUGP("No body sepration found", NULL);

  /* connect to server */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    die("socket()", strerror(errno));

  set_alarm(connect_timeout, connect_timeout_handler);
  if (connect(sock, (struct sockaddr*) &sin, sizeof(sin)))
    die("connect()", strerror(errno));
  set_alarm(0, NULL);

  /* smtp blabla */
  w2s(sock, NULL, NULL, "220");
  w2s(sock, "HELO", he->h_name, "250");
  if (from) from2socket(sock, from);

  /* look for header from address, evaluate header recips, erase bcc's */
  /* but only if we have an empty line that seperates header and body  */
  if (sep) for (left = mail.s; left < sep;) {
    /* left points to the beginning of a header line                 */
    /* right points to the end of a (possibly continued) header line */

    for (right = left;; right += 2) {
      right = strchr(right, '\r');     /* cant be NULL because sep != NULL */
      if (*(right + 2) != '\t') break; /* line is not continued            */
    }
    *right = '\0';                     /* for parsing this line            */

    if (!strncasecmp(left, "bcc:", 4)) {
      /* bcc needs special treatment */
      DEBUGP("Bcc:-line", left);
      handle_address_line(sock, left + 4, right, mode,
                          first_cmd_line_recip, &from, he->h_name,
                          &header_bcc_seen, "Bcc");
      /* erase bcc line */
      right += 2;
      memmove(left, right, mail.l - (right - mail.s));
      mail.l -= right - left;
      sep -= right - left;
    }

    else {

      /* from */
      if (!strncasecmp(left, "from:", 5)) {
        DEBUGP("From:-line", left);
        handle_address_line(sock, left + 5, right, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_from_seen, "From");
      }

      /* to */
      else if (!strncasecmp(left, "to:", 3)) {
        DEBUGP("To:-line", left);
        handle_address_line(sock, left + 3, right, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_to_seen, "To");
      }

      /* cc */
      else if (!strncasecmp(left, "cc:", 3)) {
        DEBUGP("Cc:-line", left);
        handle_address_line(sock, left + 3, right, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_cc_seen, "Cc");
      }

      /* date; no action just remember that we've seen it */
      else if (!strncasecmp(left, "date:", 5)) {
        DEBUGP("Date:-line", left);
        if (header_date_seen)
          die("your mail contains at least two Date: lines", left);
        header_date_seen = 1;
      }

      /* message-id; no action just remember that we've seen it */
      else if (!strncasecmp(left, "message-id:", 11)) {
        DEBUGP("Message-ID:-line", left);
        if (header_msgid_seen)
          die("your mail contains at least two Message-ID: lines", left);
        header_msgid_seen = 1;
      }

      /* subject; no action just remember that we've seen it */
      else if (!strncasecmp(left, "subject:", 8)) {
        DEBUGP("Message-ID:-line", left);
        if (header_subject_seen)
          die("your mail contains at least two Subject: lines", left);
        header_subject_seen = 1;
      }

      *right = '\r';    /* remove zero termination */
      left = right + 2; /* jump to next line       */
    }
  } /* for */

  /* evaluate cmd line recipients */
  if ((first_cmd_line_recip > 0) &&
      ((mode == ARGUMENT_HEADER_IF_NONE) ||
       (mode == ARGUMENT_ONLY) ||
       (mode == ARGUMENT_AND_HEADER))) {
    DEBUGP("argument recipients...", NULL);
    if (!from) {
      from = get_user(he->h_name);
      from2socket(sock, from);
    }
    for (i = first_cmd_line_recip; i < argc; i++) {
      DEBUGP("argument recipient", argv[i]);
      rcpt2socket(sock, argv[i]);
    }
  }

  /* no recipients at all? => quit program */
  if (!first_cmd_line_recip && !header_to_seen &&
      !header_cc_seen && !header_bcc_seen) 
    die("no recipients found at all", NULL);

  /* send the beast */
  str_cat_n(&mail, ".", 1);
  w2s(sock, "DATA", NULL, "354");

  /* insert unseen header lines */
  if (!header_from_seen) {
    INFOP("Generated From", from);
    w2s(sock, (*from != '<')?("From:"):("From: "), from, NULL);
  }
  if (!header_subject_seen) {
    INFOP("Generated empty Subject", NULL);
    w2s(sock, "Subject: ", "<no subject given>", NULL);
  }
  if (!header_msgid_seen) msgid2socket(sock, he->h_name);
  if (!header_date_seen) date2socket(sock);
  if (!sep) {
    INFOP("Generated body seperation", NULL);
    w2s(sock, "", NULL, NULL);
  }

  DEBUGP("sending mail...", NULL);
  w2s(sock, mail.s, NULL, "250");
  w2s(sock, "QUIT", NULL, "221");
  close(sock);

  DEBUGP("exiting...", NULL);
  return 0;
}
